unit UniCodeIniFiles;

{$R-,T-,H+,X+}

interface

{$INCLUDE TntCompilers.inc}//compiler definitions for TNT
uses SysUtils, Classes, TntClasses, {$ifndef COMPILER_9_UP}TntWideStrUtils{$else}WideStrUtils{$endif};

type
  EUniCodeIniFileException = class(Exception);

  TCustomUniCodeIniFile = class(TObject)
  private
    FFileName: WideString;
  public
    constructor Create(const FileName: WideString);
    function SectionExists(const Section: WideString): Boolean;
    function ReadWideString(const Section, Ident, Default: WideString): WideString; virtual; abstract;
    procedure WriteWideString(const Section, Ident, Value: WideString); virtual; abstract;
    function ReadInteger(const Section, Ident: WideString; Default: Longint): Longint; virtual;
    procedure WriteInteger(const Section, Ident: WideString; Value: Longint); virtual;
    function ReadBool(const Section, Ident: WideString; Default: Boolean): Boolean; virtual;
    procedure WriteBool(const Section, Ident: WideString; Value: Boolean); virtual;
    function ReadBinaryStream(const Section, Name: WideString; Value: TStream): Integer; virtual;
    function ReadDate(const Section, Name: WideString; Default: TDateTime): TDateTime; virtual;
    function ReadDateTime(const Section, Name: WideString; Default: TDateTime): TDateTime; virtual;
    function ReadFloat(const Section, Name: WideString; Default: Double): Double; virtual;
    function ReadTime(const Section, Name: WideString; Default: TDateTime): TDateTime; virtual;
    procedure WriteBinaryStream(const Section, Name: WideString; Value: TStream); virtual;
    procedure WriteDate(const Section, Name: WideString; Value: TDateTime); virtual;
    procedure WriteDateTime(const Section, Name: WideString; Value: TDateTime); virtual;
    procedure WriteFloat(const Section, Name: WideString; Value: Double); virtual;
    procedure WriteTime(const Section, Name: WideString; Value: TDateTime); virtual;
    procedure ReadSection(const Section: WideString; WideStrings: TTntStrings); virtual; abstract;
    procedure ReadSections(WideStrings: TTntStrings); virtual; abstract;
    procedure ReadSectionValues(const Section: WideString; WideStrings: TTntStrings); virtual; abstract;
    procedure EraseSection(const Section: WideString); virtual; abstract;
    procedure DeleteKey(const Section, Ident: WideString); virtual; abstract;
    procedure UpdateFile; virtual; abstract;
    function ValueExists(const Section, Ident: WideString): Boolean;
    property FileName: WideString read FFileName;
  end;

  { TWideStringHash - used internally by TMemUniCodeIniFile to optimize searches. }

  PPHashItem = ^PHashItem;
  PHashItem = ^THashItem;
  THashItem = record
    Next: PHashItem;
    Key: WideString;
    Value: Integer;
  end;

  TWideStringHash = class
  private
    Buckets: array of PHashItem;
  protected
    function Find(const Key: WideString): PPHashItem;
    function HashOf(const Key: WideString): Cardinal; virtual;
  public
    constructor Create(Size: Cardinal = 256);
    destructor Destroy; override;
    procedure Add(const Key: WideString; Value: Integer);
    procedure Clear;
    procedure Remove(const Key: WideString);
    function Modify(const Key: WideString; Value: Integer): Boolean;
    function ValueOf(const Key: WideString): Integer;
  end;

  { THashedWideStringList - A TWideStringList that uses TWideStringHash to improve the
    speed of Find }
  THashedWideStringList = class(TTntStringList)
  private
    FValueHash: TWideStringHash;
    FNameHash: TWideStringHash;
    FValueHashValid: Boolean;
    FNameHashValid: Boolean;
    procedure UpdateValueHash;
    procedure UpdateNameHash;
  protected
    procedure Changed; override;
  public
    destructor Destroy; override;
    function IndexOf(const S: WideString): Integer; override;
    function IndexOfName(const Name: WideString): Integer; override;
  end;

  { TMemUniCodeIniFile - loads an entire INI file into memory and allows all
    operations to be performed on the memory image.  The image can then
    be written out to the disk file }

  TMemUniCodeIniFile = class(TCustomUniCodeIniFile)
  private
    FSections: TTntStringList;
    function AddSection(const Section: WideString): TTntStrings;
    function GetCaseSensitive: Boolean;
    procedure LoadValues;
    procedure SetCaseSensitive(Value: Boolean);
  public
    constructor Create(const FileName: WideString);
    destructor Destroy; override;
    procedure Clear;
    procedure DeleteKey(const Section, Ident: WideString); override;
    procedure EraseSection(const Section: WideString); override;
    procedure GetStrings(List: TTntStrings);
    procedure ReadSection(const Section: WideString; WideStrings: TTntStrings); override;
    procedure ReadSections(WideStrings: TTntStrings); override;
    procedure ReadSectionValues(const Section: WideString; WideStrings: TTntStrings); override;
    function ReadWideString(const Section, Ident, Default: WideString): WideString; override;
    procedure Rename(const FileName: WideString; Reload: Boolean);
    procedure SeTTntStrings(List: TTntStrings);
    procedure UpdateFile; override;
    procedure WriteWideString(const Section, Ident, Value: WideString); override;
    property CaseSensitive: Boolean read GetCaseSensitive write SetCaseSensitive;
  end;

{$IFDEF MSWINDOWS}
  { TUniCodeIniFile - Encapsulates the Windows INI file interface
    (Get/SetPrivateProfileXXX functions) }

  TUniCodeIniFile = class(TCustomUniCodeIniFile)
  public
    destructor Destroy; override;
    function ReadWideString(const Section, Ident, Default: WideString): WideString; override;
    procedure WriteWideString(const Section, Ident, Value: WideString); override;
    procedure ReadSection(const Section: WideString; WideStrings: TTntStrings); override;
    procedure ReadSections(WideStrings: TTntStrings); override;
    procedure ReadSectionValues(const Section: WideString; WideStrings: TTntStrings); override;
    procedure EraseSection(const Section: WideString); override;
    procedure DeleteKey(const Section, Ident: WideString); override;
    procedure UpdateFile; override;
  end;
{$ELSE}
    TUniCodeIniFile = class(TMemUniCodeIniFile)
    public
      destructor Destroy; override;
    end;
{$ENDIF}


implementation

uses RTLConsts
{$IFDEF MSWINDOWS}
  , Windows
{$ENDIF};

{ TCustomUniCodeIniFile }

constructor TCustomUniCodeIniFile.Create(const FileName: WideString);
begin
  FFileName := FileName;
end;

function TCustomUniCodeIniFile.SectionExists(const Section: WideString): Boolean;
var
  S: TTntStrings;
begin
  S := TTntStringList.Create;
  try
    ReadSection(Section, S);
    Result := S.Count > 0;
  finally
    S.Free;
  end;
end;

function TCustomUniCodeIniFile.ReadInteger(const Section, Ident: WideString;
  Default: Longint): Longint;
var
  IntStr: WideString;
begin
  IntStr := ReadWideString(Section, Ident, '');
  if (Length(IntStr) > 2) and (IntStr[1] = '0') and
     ((IntStr[2] = 'X') or (IntStr[2] = 'x')) then
    IntStr := '$' + Copy(IntStr, 3, Maxint);
  Result := StrToIntDef(IntStr, Default);
end;

procedure TCustomUniCodeIniFile.WriteInteger(const Section, Ident: WideString; Value: Longint);
begin
  WriteWideString(Section, Ident, IntToStr(Value));
end;

function TCustomUniCodeIniFile.ReadBool(const Section, Ident: WideString;
  Default: Boolean): Boolean;
begin
  Result := ReadInteger(Section, Ident, Ord(Default)) <> 0;
end;

function TCustomUniCodeIniFile.ReadDate(const Section, Name: WideString; Default: TDateTime): TDateTime;
var
  DateStr: WideString;
begin
  DateStr := ReadWideString(Section, Name, '');
  Result := Default;
  if DateStr <> '' then
  try
    Result := StrToDate(DateStr);
  except
    on EConvertError do
      // Ignore EConvertError exceptions
    else
      raise;
  end;
end;

function TCustomUniCodeIniFile.ReadDateTime(const Section, Name: WideString; Default: TDateTime): TDateTime;
var
  DateStr: WideString;
begin
  DateStr := ReadWideString(Section, Name, '');
  Result := Default;
  if DateStr <> '' then
  try
    Result := StrToDateTime(DateStr);
  except
    on EConvertError do
      // Ignore EConvertError exceptions
    else
      raise;
  end;
end;

function TCustomUniCodeIniFile.ReadFloat(const Section, Name: WideString; Default: Double): Double;
var
  FloatStr: WideString;
begin
  FloatStr := ReadWideString(Section, Name, '');
  Result := Default;
  if FloatStr <> '' then
  try
    Result := StrToFloat(FloatStr);
  except
    on EConvertError do
      // Ignore EConvertError exceptions
    else
      raise;
  end;
end;

function TCustomUniCodeIniFile.ReadTime(const Section, Name: WideString; Default: TDateTime): TDateTime;
var
  TimeStr: WideString;
begin
  TimeStr := ReadWideString(Section, Name, '');
  Result := Default;
  if TimeStr <> '' then
  try
    Result := StrToTime(TimeStr);
  except
    on EConvertError do
      // Ignore EConvertError exceptions
    else
      raise;
  end;
end;

procedure TCustomUniCodeIniFile.WriteDate(const Section, Name: WideString; Value: TDateTime);
begin
  WriteWideString(Section, Name, DateToStr(Value));
end;

procedure TCustomUniCodeIniFile.WriteDateTime(const Section, Name: WideString; Value: TDateTime);
begin
  WriteWideString(Section, Name, DateTimeToStr(Value));
end;

procedure TCustomUniCodeIniFile.WriteFloat(const Section, Name: WideString; Value: Double);
begin
  WriteWideString(Section, Name, FloatToStr(Value));
end;

procedure TCustomUniCodeIniFile.WriteTime(const Section, Name: WideString; Value: TDateTime);
begin
  WriteWideString(Section, Name, TimeToStr(Value));
end;

procedure TCustomUniCodeIniFile.WriteBool(const Section, Ident: WideString; Value: Boolean);
const
  Values: array[Boolean] of WideString = ('0', '1');
begin
  WriteWideString(Section, Ident, Values[Value]);
end;

function TCustomUniCodeIniFile.ValueExists(const Section, Ident: WideString): Boolean;
var
  S: TTntStrings;
begin
  S := TTntStringList.Create;
  try
    ReadSection(Section, S);
    Result := S.IndexOf(Ident) > -1;
  finally
    S.Free;
  end;
end;

function TCustomUniCodeIniFile.ReadBinaryStream(const Section, Name: WideString;
  Value: TStream): Integer;
var
  Text: String;
  Stream: TMemoryStream;
  Pos: Integer;
begin
  Text := ReadWideString(Section, Name, '');
  if Text <> '' then
  begin
    if Value is TMemoryStream then
      Stream := TMemoryStream(Value)
    else
      Stream := TMemoryStream.Create;

    try
      Pos := Stream.Position;
      Stream.SetSize(Stream.Size + Length(Text) div 2);
      HexToBin(PChar(Text), PChar(Integer(Stream.Memory) + Stream.Position), Length(Text) div 2);
      Stream.Position := Pos;
      if Value <> Stream then
        Value.CopyFrom(Stream, Length(Text) div 2);
      Result := Stream.Size - Pos;
    finally
      if Value <> Stream then
        Stream.Free;
    end;
  end
  else
    Result := 0;
end;

procedure TCustomUniCodeIniFile.WriteBinaryStream(const Section, Name: WideString;
  Value: TStream);
var
  Text: String;
  Stream: TMemoryStream;
begin
  SetLength(Text, (Value.Size - Value.Position) * 2);
  if Length(Text) > 0 then
  begin
    if Value is TMemoryStream then
      Stream := TMemoryStream(Value)
    else
      Stream := TMemoryStream.Create;

    try
      if Stream <> Value then
      begin
        Stream.CopyFrom(Value, Value.Size - Value.Position);
        Stream.Position := 0;
      end;
      BinToHex(PChar(Integer(Stream.Memory) + Stream.Position), PChar(Text),
        Stream.Size - Stream.Position);
    finally
      if Value <> Stream then
        Stream.Free;
    end;
  end;
  WriteWideString(Section, Name, Text);
end;

{ TWideStringHash }

procedure TWideStringHash.Add(const Key: WideString; Value: Integer);
var
  Hash: Integer;
  Bucket: PHashItem;
begin
  Hash := HashOf(Key) mod Cardinal(Length(Buckets));
  New(Bucket);
  Bucket^.Key := Key;
  Bucket^.Value := Value;
  Bucket^.Next := Buckets[Hash];
  Buckets[Hash] := Bucket;
end;

procedure TWideStringHash.Clear;
var
  I: Integer;
  P, N: PHashItem;
begin
  for I := 0 to Length(Buckets) - 1 do
  begin
    P := Buckets[I];
    while P <> nil do
    begin
      N := P^.Next;
      Dispose(P);
      P := N;
    end;
    Buckets[I] := nil;
  end;
end;

constructor TWideStringHash.Create(Size: Cardinal);
begin
  inherited Create;
  SetLength(Buckets, Size);
end;

destructor TWideStringHash.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TWideStringHash.Find(const Key: WideString): PPHashItem;
var
  Hash: Integer;
begin
  Hash := HashOf(Key) mod Cardinal(Length(Buckets));
  Result := @Buckets[Hash];
  while Result^ <> nil do
  begin
    if Result^.Key = Key then
      Exit
    else
      Result := @Result^.Next;
  end;
end;

function TWideStringHash.HashOf(const Key: WideString): Cardinal;
var
  I: Integer;
begin
  Result := 0;
  for I := 1 to Length(Key) do
    Result := ((Result shl 2) or (Result shr (SizeOf(Result) * 8 - 2))) xor
      Ord(Key[I]);
end;

function TWideStringHash.Modify(const Key: WideString; Value: Integer): Boolean;
var
  P: PHashItem;
begin
  P := Find(Key)^;
  if P <> nil then
  begin
    Result := True;
    P^.Value := Value;
  end
  else
    Result := False;
end;

procedure TWideStringHash.Remove(const Key: WideString);
var
  P: PHashItem;
  Prev: PPHashItem;
begin
  Prev := Find(Key);
  P := Prev^;
  if P <> nil then
  begin
    Prev^ := P^.Next;
    Dispose(P);
  end;
end;

function TWideStringHash.ValueOf(const Key: WideString): Integer;
var
  P: PHashItem;
begin
  P := Find(Key)^;
  if P <> nil then
    Result := P^.Value
  else
    Result := -1;
end;

{ THashedWideStringList }

procedure THashedWideStringList.Changed;
begin
  inherited Changed;
  FValueHashValid := False;
  FNameHashValid := False;
end;

destructor THashedWideStringList.Destroy;
begin
  FValueHash.Free;
  FNameHash.Free;
  inherited Destroy;
end;

function THashedWideStringList.IndexOf(const S: WideString): Integer;
begin
  UpdateValueHash;
  if not CaseSensitive then
    Result :=  FValueHash.ValueOf(AnsiUpperCase(S))
  else
    Result :=  FValueHash.ValueOf(S);
end;

function THashedWideStringList.IndexOfName(const Name: WideString): Integer;
begin
  UpdateNameHash;
  if not CaseSensitive then
    Result := FNameHash.ValueOf(AnsiUpperCase(Name))
  else
    Result := FNameHash.ValueOf(Name);
end;

procedure THashedWideStringList.UpdateNameHash;
var
  I: Integer;
  P: Integer;
  Key: WideString;
begin
  if FNameHashValid then Exit;
  
  if FNameHash = nil then
    FNameHash := TWideStringHash.Create
  else
    FNameHash.Clear;
  for I := 0 to Count - 1 do
  begin
    Key := Get(I);
    P := AnsiPos('=', Key);
    if P <> 0 then
    begin
      if not CaseSensitive then
        Key := AnsiUpperCase(Copy(Key, 1, P - 1))
      else
        Key := Copy(Key, 1, P - 1);
      FNameHash.Add(Key, I);
    end;
  end;
  FNameHashValid := True;
end;

procedure THashedWideStringList.UpdateValueHash;
var
  I: Integer;
begin
  if FValueHashValid then Exit;
  
  if FValueHash = nil then
    FValueHash := TWideStringHash.Create
  else
    FValueHash.Clear;
  for I := 0 to Count - 1 do
    if not CaseSensitive then
      FValueHash.Add(AnsiUpperCase(Self[I]), I)
    else
      FValueHash.Add(Self[I], I);
  FValueHashValid := True;
end;

{ TMemUniCodeIniFile }

constructor TMemUniCodeIniFile.Create(const FileName: WideString);
begin
  inherited Create(FileName);
  FSections := THashedWideStringList.Create;
{$IFDEF LINUX}
  FSections.CaseSensitive := True;
{$ENDIF}
  LoadValues;
end;

destructor TMemUniCodeIniFile.Destroy;
begin
  if FSections <> nil then
    Clear;
  FSections.Free;
  inherited Destroy;
end;

function TMemUniCodeIniFile.AddSection(const Section: WideString): TTntStrings;
begin
  Result := THashedWideStringList.Create;
  try
    THashedWideStringList(Result).CaseSensitive := CaseSensitive;
    FSections.AddObject(Section, Result);
  except
    Result.Free;
    raise;
  end;
end;

procedure TMemUniCodeIniFile.Clear;
var
  I: Integer;
begin
  for I := 0 to FSections.Count - 1 do
    TObject(FSections.Objects[I]).Free;
  FSections.Clear;
end;

procedure TMemUniCodeIniFile.DeleteKey(const Section, Ident: WideString);
var
  I, J: Integer;
  WideStrings: TTntStrings;
begin
  I := FSections.IndexOf(Section);
  if I >= 0 then
  begin
    WideStrings := TTntStrings(FSections.Objects[I]);
    J := WideStrings.IndexOfName(Ident);
    if J >= 0 then
      WideStrings.Delete(J);
  end;
end;

procedure TMemUniCodeIniFile.EraseSection(const Section: WideString);
var
  I: Integer;
begin
  I := FSections.IndexOf(Section);
  if I >= 0 then
  begin
    TTntStrings(FSections.Objects[I]).Free;
    FSections.Delete(I);
  end;
end;

function TMemUniCodeIniFile.GetCaseSensitive: Boolean;
begin
  Result := FSections.CaseSensitive;
end;

procedure TMemUniCodeIniFile.GetStrings(List: TTntStrings);
var
  I, J: Integer;
  WideStrings: TTntStrings;
begin
  List.BeginUpdate;
  try
    for I := 0 to FSections.Count - 1 do
    begin
      List.Add('[' + FSections[I] + ']');
      WideStrings := TTntStrings(FSections.Objects[I]);
      for J := 0 to WideStrings.Count - 1 do List.Add(WideStrings[J]);
      List.Add('');
    end;
  finally
    List.EndUpdate;
  end;
end;

procedure TMemUniCodeIniFile.LoadValues;
var
  List: TTntStringList;
begin
  if (FileName <> '') and FileExists(FileName) then
  begin
    List := TTntStringList.Create;
    try
      List.LoadFromFile(FileName);
      SeTTntStrings(List);
    finally
      List.Free;
    end;
  end
  else
    Clear;
end;

procedure TMemUniCodeIniFile.ReadSection(const Section: WideString;
  WideStrings: TTntStrings);
var
  I, J: Integer;
  SectionWideStrings: TTntStrings;
begin
  WideStrings.BeginUpdate;
  try
    WideStrings.Clear;
    I := FSections.IndexOf(Section);
    if I >= 0 then
    begin
      SectionWideStrings := TTntStrings(FSections.Objects[I]);
      for J := 0 to SectionWideStrings.Count - 1 do
        WideStrings.Add(SectionWideStrings.Names[J]);
    end;
  finally
    WideStrings.EndUpdate;
  end;
end;

procedure TMemUniCodeIniFile.ReadSections(WideStrings: TTntStrings);
begin
  WideStrings.Assign(FSections);
end;

procedure TMemUniCodeIniFile.ReadSectionValues(const Section: WideString;
  WideStrings: TTntStrings);
var
  I: Integer;
begin
  WideStrings.BeginUpdate;
  try
    WideStrings.Clear;
    I := FSections.IndexOf(Section);
    if I >= 0 then
      WideStrings.Assign(TTntStrings(FSections.Objects[I]));
  finally
    WideStrings.EndUpdate;
  end;
end;

function TMemUniCodeIniFile.ReadWideString(const Section, Ident,
  Default: WideString): WideString;
var
  I: Integer;
  WideStrings: TTntStrings;
begin
  I := FSections.IndexOf(Section);
  if I >= 0 then
  begin
    WideStrings := TTntStrings(FSections.Objects[I]);
    I := WideStrings.IndexOfName(Ident);
    if I >= 0 then
    begin
      Result := Copy(WideStrings[I], Length(Ident) + 2, Maxint);
      Exit;
    end;
  end;
  Result := Default;
end;

procedure TMemUniCodeIniFile.Rename(const FileName: WideString; Reload: Boolean);
begin
  FFileName := FileName;
  if Reload then
    LoadValues;
end;

procedure TMemUniCodeIniFile.SetCaseSensitive(Value: Boolean);
var
  I: Integer;
begin
  if Value <> FSections.CaseSensitive then
  begin
    FSections.CaseSensitive := Value;
    for I := 0 to FSections.Count - 1 do
      with THashedWideStringList(FSections.Objects[I]) do
      begin
        CaseSensitive := Value;
        Changed;
      end;
    THashedWideStringList(FSections).Changed;
  end;
end;

procedure TMemUniCodeIniFile.SeTTntStrings(List: TTntStrings);
var
  I, J: Integer;
  S: WideString;
  WideStrings: TTntStrings;
begin
  Clear;
  WideStrings := nil;
  for I := 0 to List.Count - 1 do
  begin
    S := Trim(List[I]);
    if (S <> '') and (S[1] <> ';') then
      if (S[1] = '[') and (S[Length(S)] = ']') then
      begin
        Delete(S, 1, 1);
        SetLength(S, Length(S)-1);
        WideStrings := AddSection(Trim(S));
      end
      else
        if WideStrings <> nil then
        begin
          J := Pos('=', S);
          if J > 0 then // remove spaces before and after '='
            WideStrings.Add(Trim(Copy(S, 1, J-1)) + '=' + Trim(Copy(S, J+1, MaxInt)) )
          else
            WideStrings.Add(S);
        end;
  end;
end;

procedure TMemUniCodeIniFile.UpdateFile;
var
  List: TTntStringList;
begin
  List := TTntStringList.Create;
  try
    GetStrings(List);
    List.SaveToFile(FFileName);
  finally
    List.Free;
  end;
end;

procedure TMemUniCodeIniFile.WriteWideString(const Section, Ident, Value: WideString);
var
  I: Integer;
  S: WideString;
  WideStrings: TTntStrings;
begin
  I := FSections.IndexOf(Section);
  if I >= 0 then
    WideStrings := TTntStrings(FSections.Objects[I])
  else
    WideStrings := AddSection(Section);
  S := Ident + '=' + Value;
  I := WideStrings.IndexOfName(Ident);
  if I >= 0 then
    WideStrings[I] := S
  else
    WideStrings.Add(S);
end;

{$IFDEF MSWINDOWS}
{ TUniCodeIniFile }

destructor TUniCodeIniFile.Destroy;
begin
  UpdateFile;         // flush changes to disk
  inherited Destroy;
end;

function TUniCodeIniFile.ReadWideString(const Section, Ident, Default: WideString): WideString;
var
  Buffer: array[0..2047] of WideChar;
begin
  SetString(Result, Buffer, GetPrivateProfileStringW(PWideChar(Section),
    PWideChar(Ident), PWideChar(Default), Buffer, SizeOf(Buffer), PWideChar(FFileName)));
end;

procedure TUniCodeIniFile.WriteWideString(const Section, Ident, Value: WideString);
begin
  if not WritePrivateProfileStringW(PWideChar(Section), PWideChar(Ident),
                                   PWideChar(Value), PWideChar(FFileName)) then
    raise EUniCodeIniFileException.CreateResFmt(@SIniFileWriteError, [FileName]);
end;

procedure TUniCodeIniFile.ReadSections(WideStrings: TTntStrings);
const
  BufSize = 16384;
var
  Buffer, P: PWideChar;
begin
  GetMem(Buffer, BufSize);
  try
    WideStrings.BeginUpdate;
    try
      WideStrings.Clear;
      if GetPrivateProfileStringW(nil, nil, nil, Buffer, BufSize,
        PWideChar(FFileName)) <> 0 then
      begin
        P := Buffer;
        while P^ <> #0 do
        begin
          WideStrings.Add(P);
          Inc(P, WStrLen(P) + 1);
        end;
      end;
    finally
      WideStrings.EndUpdate;
    end;
  finally
    FreeMem(Buffer, BufSize);
  end;
end;

procedure TUniCodeIniFile.ReadSection(const Section: WideString; WideStrings: TTntStrings);
const
  BufSize = 16384;
var
  Buffer, P: PWideChar;
begin
  GetMem(Buffer, BufSize);
  try
    WideStrings.BeginUpdate;
    try
      WideStrings.Clear;
      if GetPrivateProfileStringW(PWideChar(Section), nil, nil, Buffer, BufSize,
        PWideChar(FFileName)) <> 0 then
      begin
        P := Buffer;
        while P^ <> #0 do
        begin
          WideStrings.Add(P);
          Inc(P, WStrLen(P) + 1);
        end;
      end;
    finally
      WideStrings.EndUpdate;
    end;
  finally
    FreeMem(Buffer, BufSize);
  end;
end;

procedure TUniCodeIniFile.ReadSectionValues(const Section: WideString; WideStrings: TTntStrings);
var
  KeyList: TTntStringList;
  I: Integer;
begin
  KeyList := TTntStringList.Create;
  try
    ReadSection(Section, KeyList);
    WideStrings.BeginUpdate;
    try
      WideStrings.Clear;
      for I := 0 to KeyList.Count - 1 do
        WideStrings.Add(KeyList[I] + '=' + ReadWideString(Section, KeyList[I], ''))
    finally
      WideStrings.EndUpdate;
    end;
  finally
    KeyList.Free;
  end;
end;

procedure TUniCodeIniFile.EraseSection(const Section: WideString);
begin
  if not WritePrivateProfileStringW(PWideChar(Section), nil, nil, PWideChar(FFileName)) then
    raise EUniCodeIniFileException.CreateResFmt(@SIniFileWriteError, [FileName]);
end;

procedure TUniCodeIniFile.DeleteKey(const Section, Ident: WideString);
begin
  WritePrivateProfileStringW(PWideChar(Section), PWideChar(Ident), nil, PWideChar(FFileName));
end;

procedure TUniCodeIniFile.UpdateFile;
begin
  WritePrivateProfileStringW(nil, nil, nil, PWideChar(FFileName));
end;
{$ELSE}

destructor TUniCodeIniFile.Destroy;
begin
  UpdateFile;
  inherited Destroy;
end;

{$ENDIF}

end.
